﻿Imports System.Collections.Specialized
Imports System.Configuration
Imports System.IO
Imports System.Text
Imports System.Threading.Tasks

Namespace SplitAndMerge
  Public Class Interpreter

'INSTANT VB NOTE: The variable instance was renamed since Visual Basic does not allow class members with the same name:
	Private Shared instance_Renamed As Interpreter

	Private Sub New()
	  Init()
	End Sub

	Public Shared ReadOnly Property Instance() As Interpreter
	  Get
		If instance_Renamed Is Nothing Then
		  instance_Renamed = New Interpreter()
		End If
		Return instance_Renamed
	  End Get
	End Property

	Private Shared ELSE_LIST As New List(Of String)()
	Private Shared ELSE_IF_LIST As New List(Of String)()

	Private CHECK_AFTER_LOOPS As Integer
	Private MAX_LOOPS As Integer

	Private m_data As String

	Private m_currentChar As Integer
	Public WriteOnly Property CurrentChar() As Integer
		Set(ByVal value As Integer)
			m_currentChar = value
		End Set
	End Property

	Private m_output As New StringBuilder()
	Public ReadOnly Property Output() As String
	  Get
'INSTANT VB NOTE: The local variable output was renamed since Visual Basic will not allow local variables with the same name as their enclosing function or property:
		Dim output_Renamed As String = m_output.ToString().Trim()
		m_output.Clear()
		Return output_Renamed
	  End Get
	End Property

	Public Sub AppendOutput(ByVal text As String, Optional ByVal newLine As Boolean = True)
	  If newLine Then
		m_output.AppendLine(text)
	  Else
		m_output.Append(text)
	  End If
	End Sub

	Public Sub Init()
	  ParserFunction.AddFunction(Constants.IF, New IfStatement(Me))
	  ParserFunction.AddFunction(Constants.WHILE, New WhileStatement(Me))
	  ParserFunction.AddFunction(Constants.BREAK, New BreakStatement())
	  ParserFunction.AddFunction(Constants.CONTINUE, New ContinueStatement())

	  ParserFunction.AddFunction(Constants.ABS, New AbsFunction())
	  ParserFunction.AddFunction(Constants.APPEND, New AppendFunction())
	  ParserFunction.AddFunction(Constants.CD, New CdFunction())
	  ParserFunction.AddFunction(Constants.CD__, New Cd__Function())
	  ParserFunction.AddFunction(Constants.DIR, New DirFunction(Me))
	  ParserFunction.AddFunction(Constants.ENV, New GetEnvFunction())
	  ParserFunction.AddFunction(Constants.EXP, New ExpFunction())
	  ParserFunction.AddFunction(Constants.FINDFILES, New FindfilesFunction(Me))
	  ParserFunction.AddFunction(Constants.FINDSTR, New FindstrFunction(Me))
	  ParserFunction.AddFunction(Constants.INDEX_OF, New IndexOfFunction())
	  ParserFunction.AddFunction(Constants.KILL, New KillFunction(Me))
	  ParserFunction.AddFunction(Constants.PI, New PiFunction())
	  ParserFunction.AddFunction(Constants.POW, New PowFunction())
	  ParserFunction.AddFunction(Constants.PRINT, New PrintFunction(Me))
	  ParserFunction.AddFunction(Constants.PSINFO, New PsInfoFunction(Me))
	  ParserFunction.AddFunction(Constants.PSTIME, New ProcessorTimeFunction())
	  ParserFunction.AddFunction(Constants.PWD, New PwdFunction(Me))
	  ParserFunction.AddFunction(Constants.RUN, New RunFunction(Me))
	  ParserFunction.AddFunction(Constants.SET, New SetVarFunction())
	  ParserFunction.AddFunction(Constants.SETENV, New SetEnvFunction())
	  ParserFunction.AddFunction(Constants.SIN, New SinFunction())
	  ParserFunction.AddFunction(Constants.SIZE, New SizeFunction())
	  ParserFunction.AddFunction(Constants.SQRT, New SqrtFunction())
	  ParserFunction.AddFunction(Constants.SUBSTR, New SubstrFunction())
	  ParserFunction.AddFunction(Constants.TOLOWER, New ToLowerFunction())
	  ParserFunction.AddFunction(Constants.TOUPPER, New ToUpperFunction())

	  ELSE_LIST.Add(Constants.ELSE)
	  ELSE_IF_LIST.Add(Constants.ELSE_IF)

	  ReadConfig()
	End Sub

	Public Sub ReadConfig()
	  CHECK_AFTER_LOOPS = ReadConfig("checkAfterLoops", 256)
	  MAX_LOOPS = ReadConfig("maxLoops", 256000)

	  Dim languagesSection = TryCast(ConfigurationManager.GetSection("Languages"), NameValueCollection)
	  If languagesSection Is Nothing OrElse languagesSection.Count = 0 Then
		Return
	  End If

	  Dim languages As String = languagesSection("languages")
	  Dim supportedLanguages() As String = languages.Split(",".ToCharArray())

	  For Each language As String In supportedLanguages
		Dim languageSection = TryCast(ConfigurationManager.GetSection(language), NameValueCollection)

		AddTranslation(languageSection, Constants.IF)
		AddTranslation(languageSection, Constants.WHILE)
		AddTranslation(languageSection, Constants.BREAK)
		AddTranslation(languageSection, Constants.CONTINUE)

		AddTranslation(languageSection, Constants.APPEND)
		AddTranslation(languageSection, Constants.CD)
		AddTranslation(languageSection, Constants.CD__)
		AddTranslation(languageSection, Constants.DIR)
		AddTranslation(languageSection, Constants.ENV)
		AddTranslation(languageSection, Constants.FINDFILES)
		AddTranslation(languageSection, Constants.FINDSTR)
		AddTranslation(languageSection, Constants.INDEX_OF)
		AddTranslation(languageSection, Constants.KILL)
		AddTranslation(languageSection, Constants.PRINT)
		AddTranslation(languageSection, Constants.PSINFO)
		AddTranslation(languageSection, Constants.PWD)
		AddTranslation(languageSection, Constants.RUN)
		AddTranslation(languageSection, Constants.SET)
		AddTranslation(languageSection, Constants.SETENV)
		AddTranslation(languageSection, Constants.SIZE)
		AddTranslation(languageSection, Constants.SUBSTR)
		AddTranslation(languageSection, Constants.TOLOWER)
		AddTranslation(languageSection, Constants.TOUPPER)

		' Special dealing for else and elif since they are not separate
		' functions but are part of the if statement block.
		AddSubstatementTranslation(languageSection, Constants.ELSE, ELSE_LIST)
		AddSubstatementTranslation(languageSection, Constants.ELSE_IF, ELSE_IF_LIST)
	  Next language
	End Sub

	Public Function ReadConfig(ByVal configName As String, Optional ByVal defaultValue As Integer = 0) As Integer
	  Dim config As String = ConfigurationManager.AppSettings(configName)
	  Dim value As Integer = defaultValue
	  If String.IsNullOrWhiteSpace(config) OrElse (Not Int32.TryParse(config, value)) Then
		Return defaultValue
	  End If

	  Return value
	End Function

	Public Sub AddTranslation(ByVal languageDictionary As NameValueCollection, ByVal originalName As String)
	  Dim translation As String = languageDictionary(originalName)
	  If String.IsNullOrWhiteSpace(translation) Then
		Return
	  End If

	  If translation.IndexOfAny((" " & vbTab & vbCrLf).ToCharArray()) >= 0 Then
		Throw New ArgumentException("Translation of [" & translation & "] contains white spaces")
	  End If

	  Dim originalFunction As ParserFunction = ParserFunction.GetFunction(originalName)
	  ParserFunction.AddFunction(translation, originalFunction)
	End Sub

	Public Sub AddSubstatementTranslation(ByVal languageDictionary As NameValueCollection, ByVal originalName As String, ByVal keywordsArray As List(Of String))
	  Dim translation As String = languageDictionary(originalName)
	  If String.IsNullOrWhiteSpace(translation) Then
		Return
	  End If

	  If translation.IndexOfAny((" " & vbTab & vbCrLf).ToCharArray()) >= 0 Then
		Throw New ArgumentException("Translation of [" & translation & "] contains white spaces")
	  End If

	  keywordsArray.Add(translation)
	End Sub

	Public Sub Process(ByVal script As String)
	  m_data = Utils.ConvertToScript(script)
	  m_currentChar = 0

	  Do While m_currentChar < m_data.Length
		Parser.LoadAndCalculate(m_data, m_currentChar, Constants.END_PARSE_ARRAY)
		Utils.GoToNextStatement(m_data, m_currentChar)
	  Loop
	End Sub

	Friend Sub ProcessWhile()
	  Dim startWhileCondition As Integer = m_currentChar

	  ' A heuristic check against an infinite loop.
	  Dim cycles As Integer = 0
	  Dim START_CHECK_INF_LOOP As Integer = CHECK_AFTER_LOOPS \ 2
	  Dim argCache1 As Parser.Result = Nothing
	  Dim argCache2 As Parser.Result = Nothing

	  Dim stillValid As Boolean = True
	  Do While stillValid
		m_currentChar = startWhileCondition

		Dim arg1 As Parser.Result = GetNextIfToken()
		Dim comparison As String = Utils.GetComparison(m_data, m_currentChar)
		Dim arg2 As Parser.Result = GetNextIfToken()

		stillValid = EvalCondition(arg1, comparison, arg2)
		Dim startSkipOnBreakChar As Integer = m_currentChar

		If Not stillValid Then
		  Exit Do
		End If

		' Check for an infinite loop if we are comparing same values:
		cycles += 1
		If cycles Mod START_CHECK_INF_LOOP = 0 Then
		  If cycles >= MAX_LOOPS OrElse (arg1.IsEqual(argCache1) AndAlso arg2.IsEqual(argCache2)) Then
			Throw New ArgumentException("Looks like an infinite loop after " & cycles & " cycles.")
		  End If
		  argCache1 = arg1
		  argCache2 = arg2
		End If

		Dim result As Parser.Result = ProcessBlock()
		If TypeOf result Is Break Then
		  m_currentChar = startSkipOnBreakChar
		  Exit Do
		End If
	  Loop

	  ' The while condition is not true anymore: must skip the whole while
	  ' block before continuing with next statements.
	  SkipBlock()
	End Sub

	Friend Function ProcessIf() As Parser.Result
	  Dim startIfCondition As Integer = m_currentChar
	  Dim result As Parser.Result = Nothing

	  Dim arg1 As Parser.Result = GetNextIfToken()
	  Dim comparison As String = Utils.GetComparison(m_data, m_currentChar)
	  Dim arg2 As Parser.Result = GetNextIfToken()

	  Dim isTrue As Boolean = EvalCondition(arg1, comparison, arg2)

	  If isTrue Then
		result = ProcessBlock()

                If TypeOf result Is [Continue] OrElse TypeOf result Is Break Then
                    ' Got here from the middle of the if-block. Skip it.
                    m_currentChar = startIfCondition
                    SkipBlock()
                End If
                SkipRestBlocks()

		Return result
	  End If

	  ' We are in Else. Skip everything in the If statement.
	  SkipBlock()

	  Dim endOfToken As Integer = m_currentChar
	  Dim nextToken As String = Utils.GetNextToken(m_data, endOfToken)

	  If ELSE_IF_LIST.Contains(nextToken) Then
		m_currentChar = endOfToken + 1
		result = ProcessIf()
	  ElseIf ELSE_LIST.Contains(nextToken) Then
		m_currentChar = endOfToken + 1
		result = ProcessBlock()
	  End If

	  Return New Parser.Result()
	End Function

	Friend Function ProcessBlock() As Parser.Result
	  Dim blockStart As Integer = m_currentChar
	  Dim result As Parser.Result = Nothing

	  Do
		Dim endGroupRead As Integer = Utils.GoToNextStatement(m_data, m_currentChar)
		If endGroupRead > 0 Then
		  Return If(result IsNot Nothing, result, New Parser.Result())
		End If

		If m_currentChar >= m_data.Length Then
		  Throw New ArgumentException("Couldn't process block [" & m_data.Substring(blockStart) & "]")
		End If

		result = Parser.LoadAndCalculate(m_data, m_currentChar, Constants.END_PARSE_ARRAY)

                If TypeOf result Is [Continue] OrElse TypeOf result Is Break Then
                    Return result
                End If
            Loop
	End Function

	Friend Sub SkipBlock()
	  Dim blockStart As Integer = m_currentChar
	  Dim startCount As Integer = 0
	  Dim endCount As Integer = 0
	  Do While startCount = 0 OrElse startCount > endCount
		If m_currentChar >= m_data.Length Then
		  Throw New ArgumentException("Couldn't skip block [" & m_data.Substring(blockStart) & "]")
		End If
		Dim currentChar As Char = m_data.Chars(m_currentChar)
		m_currentChar += 1
		Select Case currentChar
		  Case Constants.START_GROUP
			  startCount += 1
		  Case Constants.END_GROUP
			  endCount += 1
		End Select
	  Loop

	  If startCount <> endCount Then
		Throw New ArgumentException("Mismatched parentheses in data")
	  End If
	End Sub

	Friend Sub SkipRestBlocks()
	  Do While m_currentChar < m_data.Length
		Dim endOfToken As Integer = m_currentChar
		Dim nextToken As String = Utils.GetNextToken(m_data, endOfToken)
		If (Not ELSE_IF_LIST.Contains(nextToken)) AndAlso (Not ELSE_LIST.Contains(nextToken)) Then
		  Return
		End If
		m_currentChar = endOfToken
		SkipBlock()
	  Loop
	End Sub

	Friend Function EvalCondition(ByVal arg1 As Parser.Result, ByVal comparison As String, ByVal arg2 As Parser.Result) As Boolean
	  Dim compare As Boolean = If(arg1.String IsNot Nothing, CompareStrings(arg1.String, comparison, arg2.String), CompareNumbers(arg1.Value, comparison, arg2.Value))

	  Return compare
	End Function

	Friend Function CompareStrings(ByVal str1 As String, ByVal comparison As String, ByVal str2 As String) As Boolean
	  If str1 Is Nothing OrElse str2 Is Nothing Then
		Throw New ArgumentException("Inconsistent comparison data in [" & m_data.Substring(m_currentChar) & "]")
	  End If
	  Dim cmp As StringComparison = StringComparison.InvariantCultureIgnoreCase
	  Select Case comparison
		Case "=="
			Return str1.Equals(str2, cmp)
		Case "<>"
			Return Not str1.Equals(str2, cmp)
		Case "<="
			Return String.Compare(str1, str2, cmp) <= 0
		Case ">="
			Return String.Compare(str1, str2, cmp) >= 0
		Case "<"
			Return String.Compare(str1, str2, cmp) < 0
		Case ">"
			Return String.Compare(str1, str2, cmp) > 0

		Case Else
			Throw New ArgumentException("Unknown comparison: " & comparison)
	  End Select
	End Function

	Friend Function CompareNumbers(ByVal num1 As Double, ByVal comparison As String, ByVal num2 As Double) As Boolean
	  If Double.IsNaN(num1) OrElse Double.IsNaN(num2) Then
		Throw New ArgumentException("Inconsistent comparison data in [" & m_data.Substring(m_currentChar) & "]")
	  End If
	  Select Case comparison
		Case "=="
			Return num1 = num2
		Case "<>"
			Return num1 <> num2
		Case "<="
			Return num1 <= num2
		Case ">="
			Return num1 >= num2
		Case "<"
			Return num1 < num2
		Case ">"
			Return num1 > num2

		Case Else
			Throw New ArgumentException("Unknown comparison: " & comparison)
	  End Select
	End Function

	Friend Function GetNextIfToken() As Parser.Result
	  If m_currentChar >= m_data.Length OrElse m_data.Chars(m_currentChar) = Constants.END_ARG Then
		Return New Parser.Result()
	  End If

	  ' There are 2 cases: either the next token is a string, so it is
	  ' in (double) quotes, or it is a mathematical expression to be parsed.
	  If m_data.Chars(m_currentChar) = Constants.QUOTE Then
		Dim [end] As Integer = m_data.IndexOf(Constants.QUOTE, m_currentChar + 1)
		If [end] < m_currentChar Then
		  Throw New ArgumentException("Couldn't extract token from " & m_data.Substring(m_currentChar))
		End If

		m_currentChar = [end]
		Dim var As String = m_data.Substring(m_currentChar, [end] - m_currentChar)
		Return New Parser.Result(Double.NaN, var)
	  End If

	  ' Second case: parse a mathematical expression.
	  Return Parser.LoadAndCalculate(m_data, m_currentChar, Constants.COMPARE_ARRAY)
	End Function
  End Class
End Namespace
