﻿Imports System.Configuration
Imports System.IO
Imports System.Text
Imports System.Threading.Tasks
Imports Microsoft.Win32
Imports SplitAndMerge

Namespace SplitAndMerge
    ''' <summary>
    ''' Interaction logic for MainWindow.xaml
    ''' </summary>
    Partial Public Class MainWindow
        Inherits Window
        Private m_filename As String = Nothing
        Private m_baseTitle As String = Nothing
        Private m_runOnReturn As Boolean = False


        Public Sub New()
            InitializeComponent()

            textSource.Focus()
            'textSource.SelectionStart = 0;
            'textSource.SelectionLength = 0;


            m_baseTitle = Me.Title & " - "
            OpenLastIfPossible()
        End Sub

        Private Sub Open(ByVal sender As Object, ByVal e As RoutedEventArgs)
            Dim openFileDialog As New OpenFileDialog()
            openFileDialog.DefaultExt = "cscs"
            openFileDialog.Filter = "cscs files (*.cscs)|*.cscs|All files (*.*)|*.*"
            openFileDialog.FilterIndex = 0
            openFileDialog.RestoreDirectory = True

            If openFileDialog.ShowDialog() = True Then
                Try
                    m_filename = openFileDialog.FileName
                    Dim readText() As String = File.ReadAllLines(m_filename)
                    textSource.Text = String.Join(vbLf, readText)

                    Me.Title = m_baseTitle & openFileDialog.SafeFileName
                    SaveFilename(m_filename)
                Catch ex As Exception
                    MessageBox.Show("Couldn't read file from disk. Original error: " & ex.Message)
                End Try
            End If
        End Sub
        Private Sub Save(ByVal sender As Object, ByVal e As RoutedEventArgs)
            If String.IsNullOrWhiteSpace(m_filename) Then
                SaveAs(sender, e)
                Return
            End If
            SaveFile()
        End Sub

        Private Sub SaveAs(ByVal sender As Object, ByVal e As RoutedEventArgs)
            Dim saveFileDialog As New SaveFileDialog()
            saveFileDialog.DefaultExt = "cscs"
            saveFileDialog.Filter = "cscs files (*.cscs)|*.cscs|All files (*.*)|*.*"
            saveFileDialog.FilterIndex = 0
            saveFileDialog.RestoreDirectory = True

            If saveFileDialog.ShowDialog() = True Then
                m_filename = saveFileDialog.FileName
                SaveFile()

                Me.Title = m_baseTitle & saveFileDialog.SafeFileName
                SaveFilename(m_filename)
            End If
        End Sub

        Private Sub SaveFile()
            Try
                File.WriteAllText(m_filename, textSource.Text)

                Dim p As New Paragraph(New Run("File " & m_filename & " saved."))
                p.Foreground = Brushes.Black
                textResult.Document.Blocks.Add(p)
            Catch ex As Exception
                MessageBox.Show("Couldn't write file to disk. Original error: " & ex.Message)
            End Try
        End Sub

        Private Sub SaveFilename(ByVal filename As String)
            Dim config As Configuration = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None)
            config.AppSettings.Settings("lastFilename").Value = filename
            config.Save(ConfigurationSaveMode.Minimal)
            'ConfigurationManager.RefreshSection("appSettings");
        End Sub

        Private Function OpenLastFilename() As String
            Dim fileConfig = ConfigurationManager.AppSettings("lastFilename")
            If fileConfig Is Nothing Then
                Return Nothing
            End If
            Return fileConfig
        End Function

        Private Sub OpenLastIfPossible()
            m_filename = OpenLastFilename()
            If String.IsNullOrWhiteSpace(m_filename) Then
                Return
            End If

            Try
                Dim readText() As String = File.ReadAllLines(m_filename)
                textSource.Text = String.Join(vbLf, readText)

                Me.Title = m_baseTitle & System.IO.Path.GetFileName(m_filename)
                SaveFilename(m_filename)
            Catch e1 As Exception
            End Try
        End Sub

        Private Sub Clear(ByVal sender As Object, ByVal e As RoutedEventArgs)
            Dim text As String = New TextRange(textResult.Document.ContentStart, textResult.Document.ContentEnd).Text
            If String.IsNullOrWhiteSpace(text) Then
                textSource.Text = String.Empty
            Else
                textResult.Document.Blocks.Clear()
            End If
        End Sub

        Private Sub Run(ByVal sender As Object, ByVal e As RoutedEventArgs)
            'Parser.Verbose = true;

            'string script = "if(2<=1){print(2+2-(3-1));if(4>2){set(www,500);set(ww,50);print(3*www-ww*2-(-3+sin(pi/2)));print(\"IF DONE\");}}else{print(\"ELSE\");set(www,env(path));print(www);print(\"I am in ELSE\");}print(\"We are DONE.\")";
            'string script = "set(b,env(path));set(env1,\"www.ilanguage.ch\");set(env2,\"Skype\");if(indexOf(b,env2)<0){append(b,env1);}print(b);";
            'string script = "set(i,1);set(b,\"Argument\");unless(i>5){append(b,2*i);print(i);print(b);set(i,i+1);}print(\"Done!\");";

            Dim script As String = textSource.SelectedText
            If m_runOnReturn Then
                script = textSource.GetLineText(textSource.GetLineIndexFromCharacterIndex(textSource.SelectionStart - 1))
            Else
                textResult.Document.Blocks.Clear()
            End If

            If String.IsNullOrWhiteSpace(script) Then
                script = textSource.Text
            End If
            Dim errorMsg As String = Nothing
            Try
                Interpreter.Instance.Process(script)
            Catch exc As ArgumentException
                errorMsg = exc.Message
            End Try

            textResult.Foreground = Brushes.Green

            Dim output As String = Interpreter.Instance.Output
            If Not String.IsNullOrWhiteSpace(output) Then
                textResult.Document.Blocks.Add(New Paragraph(New Run(output)))
            End If
            If Not String.IsNullOrWhiteSpace(errorMsg) Then
                Dim p As New Paragraph(New Run(errorMsg))
                p.Foreground = Brushes.Red
                textResult.Document.Blocks.Add(p)
            End If
        End Sub

        Private Sub RunCheckbox(ByVal sender As Object, ByVal e As RoutedEventArgs)
            m_runOnReturn = (runOnReturn.IsChecked = True).Value
            'textSource.
        End Sub

        Private Sub TextBox_KeyEnterUpdate(ByVal sender As Object, ByVal e As KeyEventArgs)
            If m_runOnReturn AndAlso e.Key = Key.Enter Then
                Dim tBox As TextBox = CType(sender, TextBox)
                Run(sender, e)
            End If
        End Sub
    End Class
End Namespace
