﻿Imports System.Text
Imports System.Threading.Tasks

Namespace SplitAndMerge
  Public Class Parser
	Public Shared Property Verbose() As Boolean

	Public Class Result
	  Public Sub New(Optional ByVal doubleResult As Double = Double.NaN, Optional ByVal stringResult As String = Nothing, Optional ByVal tupleResult As List(Of Result) = Nothing)
		Value = doubleResult
		[String] = stringResult
		Tuple = tupleResult
	  End Sub

	  Public Sub Copy(ByVal other As Result)
		Value = other.Value
		[String] = other.String
		Tuple = other.Tuple
	  End Sub

	  Public Function IsEqual(ByVal other As Result) As Boolean
		If other Is Nothing OrElse String.IsNullOrWhiteSpace(Me.String) <> String.IsNullOrWhiteSpace(other.String) OrElse Double.IsNaN(Value) <> Double.IsNaN(other.Value) OrElse (Tuple Is Nothing) <> (other.Tuple Is Nothing) Then
		  Return False
		End If

		If (Double.IsNaN(Value) OrElse Value = other.Value) AndAlso (Me.String Is Nothing OrElse Me.String.Equals(other.String)) AndAlso (Tuple Is Nothing OrElse Tuple.SequenceEqual(other.Tuple)) Then
		  Return True
		End If

		Return False
	  End Function

	  Public Sub Reset()
		Value = Double.NaN
		[String] = Nothing
		Tuple = Nothing
	  End Sub

	  Public Property Value() As Double
	  Public Property [String]() As String
	  Public Property Tuple() As List(Of Result)
	End Class

	Private Class Cell
	  Friend Sub New(ByVal value As Double, ByVal action As Char)
		Me.Value = value
		Me.Action = action
	  End Sub

	  Friend Property Value() As Double
	  Friend Property Action() As Char
	End Class

	Public Shared Function Process(ByVal data As String) As Result
	  ' Get rid of spaces and check parenthesis
	  Dim expression As String = Preprocess(data)
	  Dim [from] As Integer = 0

	  Return LoadAndCalculate(data, [from], Constants.END_LINE_ARRAY)
	End Function

	Private Shared Function Preprocess(ByVal data As String) As String
	  If String.IsNullOrWhiteSpace(data) Then
		Throw New ArgumentException("Loaded empty data")
	  End If

	  Dim parentheses As Integer = 0
	  Dim result As New StringBuilder(data.Length)

	  For i As Integer = 0 To data.Length - 1
		Dim ch As Char = data.Chars(i)
		Select Case ch
		  Case " "c, ControlChars.Tab, ControlChars.Lf
			  Continue For
		  Case Constants.END_ARG
			  parentheses -= 1
		  Case Constants.START_ARG
			  parentheses += 1
		End Select
		result.Append(ch)
	  Next i

	  If parentheses <> 0 Then
		Throw New ArgumentException("Uneven parentheses")
	  End If

	  Return result.ToString()
	End Function

	Public Shared Function LoadAndCalculate(ByVal data As String, ByRef [from] As Integer, ByVal [to]() As Char) As Result
	  If [from] >= data.Length OrElse [to].Contains(data.Chars([from])) Then
		Return New Result()
	  End If

	  Dim listToMerge As New List(Of Cell)(16)
	  Dim item As New StringBuilder()

	  Do
	   Dim ch As Char = data.Chars([from])
	   [from] += 1
		If StillCollecting(item.ToString(), ch, [to]) Then
		  item.Append(ch)
		  If [from] < data.Length AndAlso (Not [to].Contains(data.Chars([from]))) Then
			Continue Do
		  End If
		End If

		' We are done getting the next token. The getValue() call below may
		' recursively call loadAndCalculate(). This will happen if extracted
		' item is a function or if the next item is starting with a START_ARG '('.
		Dim parsingItem As String = item.ToString()
		Dim func As New ParserFunction(data, [from], parsingItem, ch)
		Dim current As Result = func.GetValue(data, [from])

		If Double.IsNaN(current.Value) Then
		  Return current
		End If

		Dim action As Char = If(ValidAction(ch), ch, UpdateAction(data, [from], ch, [to]))

		listToMerge.Add(New Cell(current.Value, action))
		item.Clear()

	  Loop While [from] < data.Length AndAlso Not [to].Contains(data.Chars([from]))

	  If [from] < data.Length AndAlso data.Chars([from]) = Constants.END_ARG Then
		' move one char forward.
		[from] += 1
	  End If

	  Dim baseCell As Cell = listToMerge(0)
	  Dim index As Integer = 1

	  Dim result As New Result()
	  result.Value = Merge(baseCell, index, listToMerge)

	  Return result
	End Function

	Private Shared Function StillCollecting(ByVal item As String, ByVal ch As Char, ByVal [to]() As Char) As Boolean
	  If [to].Contains(ch) Then
		Return False
	  End If
	  ' Stop collecting if either got END_ARG ')' or to char, e.g. ','.
	  Dim stopCollecting As Char = If([to].Contains(Constants.END_ARG) OrElse [to].Contains(Constants.END_LINE), Constants.END_ARG, Constants.END_LINE)
	  Return (item.Length = 0 AndAlso (ch = "-"c OrElse ch = Constants.END_ARG)) OrElse Not(ValidAction(ch) OrElse ch = Constants.START_ARG OrElse ch = stopCollecting)
	End Function

	Private Shared Function ValidAction(ByVal ch As Char) As Boolean
	  Return ch = "%"c OrElse ch = "*"c OrElse ch = "/"c OrElse ch = "+"c OrElse ch = "-"c OrElse ch = "^"c
	End Function

	Private Shared Function UpdateAction(ByVal item As String, ByRef [from] As Integer, ByVal ch As Char, ByVal [to]() As Char) As Char
	  If [from] >= item.Length OrElse item.Chars([from]) = Constants.END_ARG OrElse [to].Contains(item.Chars([from])) Then
		Return Constants.END_ARG
	  End If

	  Dim index As Integer = [from]
	  Dim res As Char = ch
	  Do While (Not ValidAction(res)) AndAlso index < item.Length
		res = item.Chars(index)
		index += 1
	  Loop

	  [from] = If(ValidAction(res), index, If(index > [from], index - 1, [from]))
	  Return res
	End Function

	' From outside this function is called with mergeOneOnly = false.
	' It also calls itself recursively with mergeOneOnly = true, meaning
	' that it will return after only one merge.
	Private Shared Function Merge(ByVal current As Cell, ByRef index As Integer, ByVal listToMerge As List(Of Cell), Optional ByVal mergeOneOnly As Boolean = False) As Double
	  If Verbose Then
		PrintList(listToMerge, index - 1)
	  End If

	  Do While index < listToMerge.Count
		Dim [next] As Cell = listToMerge(index)
		index += 1

		Do While Not CanMergeCells(current, [next])
		  ' next cells first. E.g. if we have 1+2*3, we first merge next
		  ' cells, i.e. 2*3, getting 6, and then we can merge 1+6.
		  Merge([next], index, listToMerge, True) ' mergeOneOnly
		Loop
		MergeCells(current, [next])
		If mergeOneOnly Then
		  Exit Do
		End If
	  Loop

	  If Verbose Then
		Console.WriteLine("Calculated: {0}", current.Value)
	  End If
	  Return current.Value
	End Function

	Private Shared Sub MergeCells(ByVal leftCell As Cell, ByVal rightCell As Cell)
	  Select Case leftCell.Action
		Case "^"c
			leftCell.Value = Math.Pow(leftCell.Value, rightCell.Value)
		Case "%"c
			leftCell.Value = leftCell.Value Mod rightCell.Value
		Case "*"c
			leftCell.Value *= rightCell.Value
		Case "/"c
		  If rightCell.Value = 0 Then
			Throw New ArgumentException("Division by zero")
		  End If
		  leftCell.Value /= rightCell.Value
		Case "+"c
			leftCell.Value += rightCell.Value
		Case "-"c
			leftCell.Value -= rightCell.Value
	  End Select
	  leftCell.Action = rightCell.Action
	End Sub

	Private Shared Function CanMergeCells(ByVal leftCell As Cell, ByVal rightCell As Cell) As Boolean
	  Return GetPriority(leftCell.Action) >= GetPriority(rightCell.Action)
	End Function

	Private Shared Function GetPriority(ByVal action As Char) As Integer
	  Select Case action
		Case "^"c
			Return 4
		Case "%"c, "*"c, "/"c
			Return 3
		Case "+"c, "-"c
			Return 2
	  End Select
	  Return 0
	End Function

	Private Shared Sub PrintList(ByVal list As List(Of Cell), ByVal [from] As Integer)
	  Console.Write("Merging list:")
	  For i As Integer = From To list.Count - 1
		Console.Write(" ({0}, '{1}')", list(i).Value, list(i).Action)
	  Next i
	  Console.WriteLine()
	End Sub

  End Class

  Public Class ParserFunction
	Public Sub New()
	  m_impl = Me
	End Sub

	' A "virtual" Constructor
	Friend Sub New(ByVal data As String, ByRef [from] As Integer, ByVal item As String, ByVal ch As Char)
	  If item.Length = 0 AndAlso ch = Constants.START_ARG Then
		' There is no function, just an expression in parentheses
		m_impl = s_idFunction
		Return
	  End If

	  m_impl = GetFunction(item)
	End Sub

	Public Shared Function GetFunction(ByVal item As String) As ParserFunction
            Dim impl As ParserFunction = Nothing
	  If s_functions.TryGetValue(item, impl) Then
		' Function exists and is registered (e.g. pi, exp, etc.)
		Return impl
	  End If

	  ' Function not found, will try to parse this as a number.
	  s_strtodFunction.Item = item
	  impl = s_strtodFunction
	  Return impl
	End Function

	Public Shared Function FunctionExists(ByVal item As String) As Boolean
	  Return s_functions.ContainsKey(item)
	End Function

	Public Shared Sub AddFunction(ByVal name As String, ByVal [function] As ParserFunction)
	  s_functions(name) = [function]
	End Sub

	Public Function GetValue(ByVal data As String, ByRef [from] As Integer) As Parser.Result
	  Return m_impl.Evaluate(data, [from])
	End Function

	Protected Overridable Function Evaluate(ByVal data As String, ByRef [from] As Integer) As Parser.Result
	  ' The real implementation will be in the derived classes.
	  Return New Parser.Result()
	End Function

	Private m_impl As ParserFunction
	Private Shared s_functions As New Dictionary(Of String, ParserFunction)()

	Private Shared s_strtodFunction As New StrtodFunction()
	Private Shared s_idFunction As New IdentityFunction()
  End Class

  Friend Class StrtodFunction
	  Inherits ParserFunction
	Protected Overrides Function Evaluate(ByVal data As String, ByRef [from] As Integer) As Parser.Result
	  Dim num As Double
	  If Not Double.TryParse(Item, num) Then
		Throw New ArgumentException("Couldn't parse token [" & Item & "]")
	  End If
	  Return New Parser.Result(num)
	End Function
	Private privateItem As String
	Public Property Item() As String
		Private Get
			Return privateItem
		End Get
		Set(ByVal value As String)
			privateItem = value
		End Set
	End Property
  End Class

  Friend Class IdentityFunction
	  Inherits ParserFunction
	Protected Overrides Function Evaluate(ByVal data As String, ByRef [from] As Integer) As Parser.Result
	  Return Parser.LoadAndCalculate(data, [from], Constants.END_ARG_ARRAY)
	End Function
  End Class

End Namespace
