﻿Imports System.IO
Imports System.Text
Imports System.Threading
Imports System.Threading.Tasks

Namespace SplitAndMerge
  Public Class Utils
	Public Shared Function GetItem(ByVal data As String, ByRef [from] As Integer, Optional ByVal expectInt As Boolean = False) As Parser.Result
	  Dim value As New Parser.Result()
	  If data.Length <= From Then
		Throw New ArgumentException("Incomplete function definition")
	  End If
	  If data.Chars([from]) = Constants.NEXT_ARG Then
		[from] += 1
	  End If

	  ' Utils.GoToNextStatement(data, ref from);
	  If [from] < data.Length AndAlso data.Chars([from]) = Constants.QUOTE Then
		[from] += 1 ' skip first quote
		If [from] < data.Length AndAlso data.Chars([from]) = Constants.QUOTE Then
		  value.String = ""
		Else
		  value.String = Utils.GetToken(data, [from], Constants.QUOTE_ARRAY)
		End If
		[from] += 1 ' skip next separation char
	  Else
		Dim existing As Parser.Result = Parser.LoadAndCalculate(data, [from], Constants.NEXT_OR_END_ARRAY)
		value.Copy(existing)
	  End If

	  If expectInt AndAlso Double.IsNaN(value.Value) Then
		Throw New ArgumentException("Integer expected instead of [" & value.String & "]")
	  End If

	  If [from] < data.Length AndAlso data.Chars([from]) = Constants.END_ARG Then
		[from] += 1
	  End If
	  Return value
	End Function

	Public Shared Function GetToken(ByVal data As String, ByRef [from] As Integer, ByVal [to]() As Char) As String
	  If data.Length <= [from] + 1 Then
		Throw New ArgumentException("End of line before extracting token")
	  End If
	  If data.Chars([from]) = Constants.QUOTE Then
		[from] += 1
	  End If

	  Dim [end] As Integer = data.IndexOfAny([to], [from])

	  If [from] = [end] Then
		[from] += 1
		Return String.Empty
	  End If

	  ' Skip found characters that have a backslash before.
	  Do While ([end] > 0 AndAlso data.Chars([end] - 1) = "\"c) AndAlso [end] + 1 < data.Length
		[end] = data.IndexOfAny([to], [end] + 1)
	  Loop

	  If [end] < From Then
		Throw New ArgumentException("Couldn't extract token from " & data.Substring([from]))
	  End If

	  If data.Chars([end] - 1) = Constants.QUOTE Then
		[end] -= 1
	  End If

	  Dim var As String = data.Substring([from], [end] - [from])
	  ' \"yes\" --> "yes"
	  var = var.Replace("\""", """")
	  [from] = [end] + 1

	  Return var
	End Function

	Public Shared Function GetNextToken(ByVal data As String, ByRef [from] As Integer) As String
	  If [from] >= data.Length Then
		Return ""
	  End If

	  Dim [end] As Integer = data.IndexOfAny(Constants.TOKEN_SEPARATION, [from])

	  If [end] < From Then
		Return ""
	  End If

	  Dim var As String = data.Substring([from], [end] - [from])
	  [from] = [end]
	  Return var
	End Function

	Public Shared Function GoToNextStatement(ByVal data As String, ByRef [from] As Integer) As Integer
	  Dim endGroupRead As Integer = 0
	  Do While [from] < data.Length
		Dim currentChar As Char = data.Chars([from])
		Select Case currentChar
		  Case Constants.END_GROUP
			  endGroupRead += 1
			[from] += 1
			Return endGroupRead
		  Case Constants.START_GROUP, Constants.QUOTE, Constants.END_STATEMENT, Constants.END_ARG
			  [from] += 1
		  Case Else
			  Return endGroupRead
		End Select
	  Loop
	  Return endGroupRead
	End Function

	Public Shared Function GetComparison(ByVal data As String, ByRef [from] As Integer) As String
	  If data.Length <= [from] + 1 Then
		Throw New ArgumentException("End of line before extracting comparison token")
	  End If

	  If Not Utils.IsCompareSign(data.Chars([from] + 1)) Then
                If data.Chars([from]) = "<"c OrElse data.Chars([from]) = ">"c Then
                    Dim result1 As String = data.Substring([from], 1)
                    [from] += 1
                    Return result1
                End If
                Throw New ArgumentException("Couldn't extract comparison token from " & data.Substring([from]))
	  End If

	  Dim result As String = data.Substring([from], 2)
	  [from] += 1
	  [from] += 1
	  Return result
	End Function

	Public Shared Function IsCompareSign(ByVal ch As Char) As Boolean
	  Return ch = "<"c OrElse ch = ">"c OrElse ch = "="c
	End Function

	Public Shared Function IsAndOrSign(ByVal ch As Char) As Boolean
	  Return ch = "&"c OrElse ch = "|"c
	End Function

	' Checks whether there is an argument separator (e.g.  ',') before the end of the
	' function call. E.g. returns true for "a,b)" and "a(b,c),d)" and false for "b),c".
	Public Shared Function SeparatorExists(ByVal data As String, ByVal [from] As Integer) As Boolean
	  If [from] >= data.Length Then
		Return False
	  End If

	  Dim argumentList As Integer = 0
	  For i As Integer = From To data.Length - 1
		Dim ch As Char = data.Chars(i)
		Select Case ch
		  Case Constants.NEXT_ARG
			Return True
		  Case Constants.START_ARG
			argumentList += 1
		  Case Constants.END_STATEMENT, Constants.END_GROUP, Constants.END_ARG
			argumentList -= 1
			If argumentList < 0 Then
			  Return False
			End If
		End Select
	  Next i

	  Return False
	End Function

	Public Shared Function GetFunctionArgs(ByVal data As String, ByRef [from] As Integer) As List(Of String)
	  Dim args As New List(Of String)()
	  Dim moreArgs As Boolean = True
	  Do While moreArgs
		moreArgs = Utils.SeparatorExists(data, [from])
		Dim item As Parser.Result = Utils.GetItem(data, [from])

		' Separate treatment for an array.
		' Only 1-dimensional arrays are supported at the moment.
		If item.Tuple IsNot Nothing AndAlso item.Tuple.Count > 0 Then
		  For i As Integer = 0 To item.Tuple.Count - 1
			Dim arg As Parser.Result = item.Tuple(i)
			args.Add((If(arg.String IsNot Nothing, arg.String, arg.Value.ToString())) & ControlChars.Lf)
		  Next i
		  Continue Do
		End If

		If item.String Is Nothing AndAlso Double.IsNaN(item.Value) Then
		  Exit Do
		End If
		args.Add(If(item.String IsNot Nothing, item.String, item.Value.ToString()))
	  Loop

	  Return args
	End Function

	Public Shared Function ExtractArrayElement(ByRef varName As String) As Integer
	  Dim argStart As Integer = varName.IndexOf(Constants.START_ARG)
	  If argStart <= 0 Then
		Return -1
	  End If

	  Dim argEnd As Integer = varName.IndexOf(Constants.END_ARG, argStart + 1)
	  If argEnd <= argStart + 1 Then
		Return -1
	  End If

	  Dim getIndexFrom As Integer = argStart
	  Dim existing As Parser.Result = Parser.LoadAndCalculate(varName, getIndexFrom, Constants.NEXT_OR_END_ARRAY)

	  If (Not Double.IsNaN(existing.Value)) AndAlso existing.Value >= 0 Then
		varName = varName.Substring(0, argStart)
		Return CInt(Fix(existing.Value))
	  End If

	  Return -1
	End Function

	Public Shared Function ConvertToScript(ByVal source As String) As String
	  Dim sb As New StringBuilder(source.Length)
	  Dim inQuotes As Boolean = False
	  Dim inComments As Boolean = False
	  Dim previous As Char = ControlChars.NullChar

	  Dim parentheses As Integer = 0
	  Dim groups As Integer = 0

	  For i As Integer = 0 To source.Length - 1
		Dim ch As Char = source.Chars(i)
		Dim [next] As Char = If(i + 1 < source.Length, source.Chars(i + 1), ControlChars.NullChar)
		Select Case ch
		  Case "/"c
			If inComments OrElse [next] = "/"c Then
			  inComments = True
			  Continue For
			End If
		  Case """"c
			If Not inComments Then
			  If previous <> "\"c Then
				  inQuotes = Not inQuotes
			  End If
			End If
		  Case " "c
			If inQuotes Then
				sb.Append(ch)
			End If
			Continue For
		  Case ControlChars.Tab
			If inQuotes Then
				sb.Append(ch)
			End If
			Continue For
		  Case ControlChars.Cr, ControlChars.Lf
			inComments = False
			Continue For
		  Case Constants.END_ARG
			parentheses -= 1
		  Case Constants.START_ARG
			parentheses += 1
		  Case Constants.END_GROUP
			  groups -= 1
		  Case Constants.START_GROUP
			  groups += 1
		  Case Else
		End Select
		If Not inComments Then
		  sb.Append(ch)
		End If
		previous = ch
	  Next i

	  If parentheses <> 0 Then
		Throw New ArgumentException("Uneven parentheses " & Constants.START_ARG + Constants.END_ARG)
	  End If
	  If groups <> 0 Then
		Throw New ArgumentException("Uneven groups " & Constants.START_GROUP + Constants.END_GROUP)
	  End If

	  Return sb.ToString()
	End Function

	Public Shared Function GetFiles(ByVal path As String, ByVal patterns() As String) As List(Of String)
	  Try
                Dim files As List(Of String) = patterns.SelectMany(
                    Function(i) Directory.EnumerateFiles(path, i)
                    ).ToList()

                Parallel.ForEach(Directory.EnumerateDirectories(path), Sub(dir)
			Dim extraFiles As List(Of String) = GetFiles(dir, patterns)
			If extraFiles IsNot Nothing AndAlso extraFiles.Count > 0 Then
				SyncLock s_mutexLock
					files.AddRange(extraFiles)
				End SyncLock
			End If
		End Sub)

		Return files
	  Catch e1 As Exception
		Return Nothing
	  End Try
	End Function

	Public Shared Function GetStringInFiles(ByVal path As String, ByVal search As String, ByVal patterns() As String, Optional ByVal ignoreCase As Boolean = True) As List(Of String)
	  Dim allFiles As List(Of String) = GetFiles(path, patterns)
	  Dim results As New List(Of String)()

	  If allFiles Is Nothing AndAlso allFiles.Count = 0 Then
		Return results
	  End If

	  Dim caseSense As StringComparison = If(ignoreCase, StringComparison.OrdinalIgnoreCase, StringComparison.Ordinal)
	  Parallel.ForEach(allFiles, Sub(currentFile)
		  Dim contents As String = GetFileText(currentFile)
		  If contents.IndexOf(search, caseSense) >= 0 Then
			  SyncLock s_mutexLock
				  results.Add(currentFile)
			  End SyncLock
		  End If
	  End Sub)

	  Return results
	End Function

	Public Shared Function GetFileText(ByVal name As String) As String
	  Dim fileContents As String = String.Empty
	  If File.Exists(name) Then
		fileContents = File.ReadAllText(name)
	  End If
	  Return fileContents
	End Function

	Public Shared Function GetLine(Optional ByVal chars As Integer = 75) As String
	  Return String.Format("-").PadRight(chars, "-"c)
	End Function

	Private Shared ReadOnly s_mutexLock As New Object()
  End Class
End Namespace
